/* main.c - Application main entry point */


#include <zephyr/types.h>
#include <stddef.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/util.h>
#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/drivers/i2c.h>
#include <zephyr/pm/device.h>
#include <zephyr/sys/poweroff.h>
#include <soc.h>
#include <stdio.h>
#include <string.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <bluetooth/services/nus.h>
#include <zephyr/settings/settings.h>
#include <version.h>

#include <zephyr/drivers/flash.h>
#include <zephyr/storage/flash_map.h>

#include <zephyr/debug/thread_analyzer.h>

#include <zephyr/drivers/clock_control/nrf_clock_control.h>
/*
#include <nrfx_gpiote.h>
#include <helpers/nrfx_gppi.h>
#if defined(DPPI_PRESENT)
#include <nrfx_dppi.h>
#else
#include <nrfx_ppi.h>
#endif
*/

#include <nrfx_lpcomp.h>
#include "nrfx_uarte.h"

#include "retained.h"
#include <hal/nrf_gpio.h>
#include <hal/nrf_power.h>
#include <nrfx_nvmc.h>
#include <nrfx_timer.h>

#include "imu/inv_imu_driver.h"
#include "gyro_helper.h"
#include <nrfx_rtc.h>
#include "analog_in.h"

#include "adv_config.h"
#include "freq_meter.h"
#include "main.h"
#include "compile_timestamp.h"

#define RTC_INSTANCE 2
#define CC_CHANNEL 0
#define RTC_TIMEOUT_MS 1000
#define RTC_PRESCALER_MS 125

#define STORAGE_PARTITION_OFFSET FIXED_PARTITION_OFFSET(storage_partition)
#define STORAGE_PARTITION_DEVICE FIXED_PARTITION_DEVICE(storage_partition)
#define FLASH_PAGE_SIZE 4096

#define DEBUG_PRINT 1

#define UARTE_INST_IDX 0

nrfx_rtc_t rtc = NRFX_RTC_INSTANCE(RTC_INSTANCE);
nrfx_uarte_t uarte_instance = NRFX_UARTE_INSTANCE(0);

static uint8_t uarte_tx_buffer[UART_TX_BUFFER_SIZE];
static uint8_t uarte_rx_buffer[UART_RX_BUFFER_SIZE];

static K_SEM_DEFINE(ble_init_ok, 0, 1);

static struct bt_conn *current_conn;
static struct bt_conn *auth_conn;

uint8_t sensor_data[SENSOR_DATA_LEN];

static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
	//	BT_DATA_BYTES(BT_DATA_UUID16_ALL, BT_UUID_16_ENCODE(BT_UUID_DIS_VAL)),
	BT_DATA(BT_DATA_MANUFACTURER_DATA, sensor_data, SENSOR_DATA_LEN),
};

static const struct bt_data sd[] = {
	BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_NUS_VAL),
};
struct uart_data_tx_t
{
	void *fifo_reserved;
	uint8_t data[UART_TX_FIFO_BUFFER_SIZE];
	uint32_t len;
};
struct uart_data_rx_t
{
	void *fifo_reserved;
	uint8_t data[UART_RX_FIFO_BUFFER_SIZE];
	uint32_t len;
};
K_FIFO_DEFINE(message_fifo);
K_FIFO_DEFINE(rx_fifo);
K_FIFO_DEFINE(tx_fifo);

struct message_item_t
{
	void *fifo_reserved;
	uint8_t message[SENSOR_DATA_LEN];
};
size_t message_item_size = sizeof(struct message_item_t);

nrfx_err_t nrfx_err;
nrfx_lpcomp_config_t lpcomp_config; // = NRFX_LPCOMP_DEFAULT_CONFIG(NRF_LPCOMP_INPUT_2);

#define INPUT_BUFFER_LEN 128

static char str[INPUT_BUFFER_LEN];
static inv_imu_sensor_event_t imu_event;

#if CONFIG_BT_NUS_UART_ASYNC_ADAPTER
UART_ASYNC_ADAPTER_INST_DEFINE(async_adapter);
#else
//static const struct device *const async_adapter;
#endif

const struct device *flash_dev = STORAGE_PARTITION_DEVICE;

volatile uint32_t timeout = 0;
volatile uint8_t irq_received = 0;
volatile uint32_t tilt_counter = 0;
volatile uint32_t in_counter = 0;
volatile uint32_t tilt_counter_th = 0;
volatile uint32_t in_counter_th = 0;
volatile uint32_t total_tilt_counter = 0;
volatile uint32_t total_in_counter = 0;
volatile uint8_t uarte_mode = UARTE_MODE_RS232;
volatile uint8_t uarte_new_data = 0;
//volatile uint8_t rx_in_progress = 0, tx_in_progress = 0;
volatile int fifo_ret = 0;
volatile uint8_t malloc_error = 0;
volatile uint16_t tx_queue_items = 0;
volatile bool fifo_ok = true;
volatile bool main_ready = false;

uint8_t config_in_progress = 0;
uint8_t uarte_initialized = 0;

uint8_t rtc_initialized = 0;
nrfx_rtc_config_t rtc_config = NRFX_RTC_DEFAULT_CONFIG;

typedef void (*tilt_irq_handler_t)(void);
tilt_irq_handler_t tilt_handler;
typedef void (*in_irq_handler_t)(void);
in_irq_handler_t in_handler;

K_SEM_DEFINE(tx_fifo_sem, 1, 1);
K_SEM_DEFINE(rx_fifo_sem, 1, 1);
K_SEM_DEFINE(message_fifo_sem, 1, 1);
K_SEM_DEFINE(gyro_sem, 1, 1);
K_SEM_DEFINE(analog_sem, 1, 1);
K_SEM_DEFINE(debug_message_sem, 1, 1);
K_SEM_DEFINE(uarte_rx_sem, 0, 1);

uint8_t accel_initialized = 0;
uint8_t gyro_initialized = 0;
uint8_t rs232_initialized = 0;
uint8_t rs485_initialized = 0;
uint32_t mac_address[2], serial_number;

void counter_timers_init(void);

union flash_buf
{
	uint32_t uintArray[10];
	float floatArray[10];
} calibration_data;

union rs_readings {
	uint32_t uint_val;
	float float_val;
} rs_readings_union;

static int sensor_configuration_settings_set(const char *name, size_t len, settings_read_cb read_cb, void *cb_arg)
{
	const char *next;
	int rc;
	if (settings_name_steq(name, "sensor", &next) && !next)
	{
		if (len != sizeof(retained))
		{
			return -EINVAL;
		}

		rc = read_cb(cb_arg, &retained, sizeof(retained));
		if (rc >= 0)
		{
			return 0;
		}

		return rc;
	}
	return -ENOENT;
}

struct settings_handler sensor_conf = {
	.name = "conf",
	.h_set = sensor_configuration_settings_set};

void gyro_irq_handler(void)
{
	irq_received = 1;
	timeout = retained.timeout;
}
/*
void tilt_irq_handler(void)
{
	tilt_counter++;
	total_tilt_counter++;
	timeout=retained.timeout;
}

void tilt_callback(const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{
	if (tilt_handler == NULL)
	{
		return;
	}

	tilt_handler();
}

void in_irq_handler(void)
{
	in_counter++;
	total_in_counter++;
	timeout=retained.timeout;
}

void in_callback(const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{
	if (in_handler == NULL)
	{
		return;
	}

	in_handler();
}
*/

static void comparator_handler(nrf_lpcomp_event_t event)
{
	if (retained.debug)
	{
		// printk("LPCOMP event");
	}
}

void suspend_threads(void);
void resume_configured_threads(void);

void uarte_callback(nrfx_uarte_event_t const *p_event, void *p_context)
{
	struct uart_data_tx_t *tx_item;	
	// struct uart_data_rx_t rx_item;

	switch (p_event->type)
	{
	case NRFX_UARTE_EVT_TX_DONE:					
		if (!config_in_progress && k_fifo_is_empty(&tx_fifo)) // TODO - Here verify
		{			
			nrfx_uarte_uninit(&uarte_instance);
			uarte_initialized = 0;
			if (uarte_mode == UARTE_MODE_RS485)
			{
				nrf_gpio_pin_clear(UARTE_RS485_RXTX_PIN);
			}		
		}				
		if(!k_fifo_is_empty(&tx_fifo)) {
			if(k_sem_take(&tx_fifo_sem, K_NO_WAIT)) {
				fifo_ok = false;
			} else {
				fifo_ok = true;
				tx_item = k_fifo_get(&tx_fifo,K_NO_WAIT);
				tx_queue_items--;
				k_sem_give(&tx_fifo_sem);
				if(tx_item != NULL){ // && tx_item->len > 0 && tx_item->len < sizeof(uarte_tx_buffer)) {
					memcpy(uarte_tx_buffer,tx_item->data,tx_item->len);					
					nrfx_uarte_tx(&uarte_instance, uarte_tx_buffer, tx_item->len, 0);
					k_free(tx_item);
				}
			}
		} else {
			if(uarte_mode==UARTE_MODE_RS485) {
				nrf_gpio_pin_clear(UARTE_RS485_RXTX_PIN);
			}			
		}				
		break;
#if (KERNEL_VERSION_MAJOR == 3) && (KERNEL_VERSION_MINOR >= 6)		
	case NRFX_UARTE_EVT_RX_BUF_REQUEST:
		break;
	case NRFX_UARTE_EVT_RX_DISABLED:		
		break;
	case NRFX_UARTE_EVT_RX_BUF_TOO_LATE:
		break;
	case NRFX_UARTE_EVT_TRIGGER:
		break;				

	case NRFX_UARTE_EVT_RX_BYTE:
		uarte_new_data = 1; // p_event->data.rx.length;
		break;
	case NRFX_UARTE_EVT_RX_DONE:	
		uarte_new_data = p_event->data.rx.length;
		k_sem_give(&uarte_rx_sem);
		break;	
#else
	case NRFX_UARTE_EVT_RX_DONE:	
			uarte_new_data = p_event->data.rx.bytes;
			break;		
#endif		
	
		
		
	case NRFX_UARTE_EVT_ERROR:		
		uarte_new_data = 0;
		k_sem_give(&uarte_rx_sem);
		break;
	}
}

void uarte_init(uint8_t mode, uint32_t baud)
{
//#ifndef CONFIG_SERIAL
	nrfx_uarte_config_t uarte_config = NRFX_UARTE_DEFAULT_CONFIG(UARTE_RS232_TX_PIN, UARTE_RS232_RX_PIN);
	if (uarte_initialized)
	{
		nrfx_uarte_tx_abort(&uarte_instance, true);
		nrfx_uarte_rx_abort(&uarte_instance, true, true);
		nrfx_uarte_uninit(&uarte_instance);
		uarte_initialized = 0;
		uarte_new_data = 0;
	}
	uarte_config.cts_pin = NRF_UARTE_PSEL_DISCONNECTED;
	uarte_config.rts_pin = NRF_UARTE_PSEL_DISCONNECTED;
	uarte_mode = mode;
	switch(baud) {
		case 1200: 
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud1200;
			break;
		case 2400:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud2400;
			break;
		case 4800:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud4800;
			break;
		case 9600:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud9600;
			break;
		case 14400:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud14400;
			break;
		case 19200:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud19200;
			break;
		case 28800:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud28800;
			break;
		case 38400:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud38400;
			break;
		case 57600:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud57600;
			break;
		case 76800:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud76800;
			break;
		case 115200:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud115200;
			break;
		case 230400:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud230400;
			break;
		case 250000:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud250000;
			break;
		case 460800:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud460800;
			break;
		case 921600:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud921600;
			break;
		case 1000000:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud1M;
			break;
		default:
			uarte_config.baudrate = UARTE_BAUDRATE_BAUDRATE_Baud115200;
			break;
	}	
	uarte_config.p_context = &uarte_instance;
	if (mode == UARTE_MODE_RS232)
	{
		uarte_config.rxd_pin = UARTE_RS232_RX_PIN;
		uarte_config.txd_pin = UARTE_RS232_TX_PIN;
	}
	else if (mode == UARTE_MODE_RS485)
	{
		uarte_config.rxd_pin = UARTE_RS485_RX_PIN;
		uarte_config.txd_pin = UARTE_RS485_TX_PIN;
		nrf_gpio_cfg_output(UARTE_RS485_RXTX_PIN);
		nrf_gpio_pin_clear(UARTE_RS485_RXTX_PIN);
	}
	else
	{
		return;
	}
#if defined(__ZEPHYR__)
	IRQ_DIRECT_CONNECT(NRFX_IRQ_NUMBER_GET(NRF_UARTE_INST_GET(UARTE_INST_IDX)), IRQ_PRIO_LOWEST,
					   NRFX_UARTE_INST_HANDLER_GET(UARTE_INST_IDX), 0);
#endif
	if (nrfx_uarte_init(&uarte_instance, &uarte_config, uarte_callback) == NRFX_SUCCESS)
	{ //
		uarte_initialized = 1;
	}
	if (config_in_progress)
	{		
#if (KERNEL_VERSION_MAJOR == 3) && (KERNEL_VERSION_MINOR >= 6)		
		nrfx_uarte_rxdrdy_enable(&uarte_instance);
#endif		
		nrfx_uarte_rx(&uarte_instance, uarte_rx_buffer, sizeof(uarte_rx_buffer));		
	}	
//#else
//#warning "uarte_init() disabled in config!"
//#endif
}

void uart_start_tx(void)
{
	struct uart_data_tx_t *uart_data;
	k_sem_take(&tx_fifo_sem, K_FOREVER);
	uart_data = k_fifo_get(&tx_fifo, K_NO_WAIT);
	tx_queue_items--;
	k_sem_give(&tx_fifo_sem);
	if (uart_data != NULL)
	{
		if(!uarte_initialized){
			uarte_init(UARTE_MODE_RS232,retained.uart_baud);	
		}
		if (uarte_mode == UARTE_MODE_RS485)
		{
			nrf_gpio_pin_set(UARTE_RS485_RXTX_PIN);
		}		
		memcpy(uarte_tx_buffer, uart_data->data, uart_data->len);
		nrfx_uarte_tx(&uarte_instance, uarte_tx_buffer, uart_data->len, 0);//NRFX_UARTE_TX_BLOCKING);
		k_free(uart_data);
	}
}

void uarte_send(uint8_t *data, size_t length)
{
//#ifndef CONFIG_SERIAL
	struct uart_data_tx_t *tx_item;
	size_t tx_item_size = sizeof(struct uart_data_tx_t);
	size_t chunk_size = sizeof(tx_item->data);
	size_t offset = 0;
	//if (uarte_initialized)
	//{
		while (offset < length)
		{
			tx_item = k_malloc(tx_item_size);
			if (tx_item == NULL)
        	{
            	malloc_error = 1;
            	// Możesz dodać dodatkowe kroki obsługi błędu tutaj, np. logowanie błędu
            	break;  // Przerwij pętlę, ponieważ alokacja pamięci się nie powiodła
        	}
			tx_item->len = (length - offset < chunk_size) ? (length - offset) : chunk_size;
			memcpy(tx_item->data, data + offset, tx_item->len);
			k_sem_take(&tx_fifo_sem, K_FOREVER);
			k_fifo_put(&tx_fifo, tx_item);
			tx_queue_items++;
			k_sem_give(&tx_fifo_sem);
			offset += tx_item->len;
		}
		if(!uarte_initialized){
			uarte_init(UARTE_MODE_RS232,retained.uart_baud);				
		}
		if (!nrfx_uarte_tx_in_progress(&uarte_instance))
		{
			uart_start_tx();
		}
		// k_yield();
	//}
//#endif
}

/*
void uart_send_task(void *p1, void *p2, void *p3)
{
	while(!uarte_initialized) {
		k_msleep(5);
	}
	while (1)
	{
		if (!k_fifo_is_empty(&tx_fifo) && !nrfx_uarte_tx_in_progress(&uarte_instance))
		{
			if (!uarte_initialized)
			{
				uarte_init(UARTE_MODE_RS232,retained.uart_baud);
			}
			uart_start_tx();
		} else {
			k_msleep(250);
		}
	}
}
*/
// K_TIMER_DEFINE(uart_send_timer, uart_send_handler, NULL);
//K_THREAD_DEFINE(th_uart, 512, uart_send_task, NULL, NULL, NULL, UART_SEND_PRIORITY, 0, 0);

void my_printk(uint8_t *data)
{
	uint8_t len;
//#ifndef CONFIG_SERIAL
	if (retained.debug)
	{
		len = strlen(data);
		uarte_send(data, len);
	}
//#else
//	printk("%s", data);
//#endif
}

static void connected(struct bt_conn *conn, uint8_t err)
{
	char addr[BT_ADDR_LE_STR_LEN];

	if (err)
	{
#if DEBUG_PRINT
		sprintf(str, "Connection failed (err %u)\r\n", err);
		my_printk(str);
#endif
		return;
	}

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
#if DEBUG_PRINT
	sprintf(str, "Connected %s\r\n", addr);
	my_printk(str);
#endif

	current_conn = bt_conn_ref(conn);
	config_in_progress = 1;
	if (!uarte_initialized)
	{
		uarte_init(UARTE_MODE_RS232,retained.uart_baud);
	}
	suspend_threads();
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
#if DEBUG_PRINT
	sprintf(str, "Disconnected: %s (reason %u)\r\n", addr, reason);
	my_printk(str);
#endif

	if (auth_conn)
	{
		bt_conn_unref(auth_conn);
		auth_conn = NULL;
	}

	if (current_conn)
	{
		bt_conn_unref(current_conn);
		current_conn = NULL;
	}
	config_in_progress = 0;
	timeout = retained.timeout;
	retained_update();
	settings_save_one("conf/sensor", &retained, sizeof(retained));
#if DEBUG_PRINT
	sprintf(str, "Debug: ts: %d\r\n", retained.config_timestamp);
	my_printk(str);
#endif
	resume_configured_threads();
}

#ifdef CONFIG_BT_NUS_SECURITY_ENABLED
static void security_changed(struct bt_conn *conn, bt_security_t level,
							 enum bt_security_err err)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
	/*
		if (!err) {
			printk("Security changed: %s level %u", addr, level);
		} else {
			printk("Security failed: %s level %u err %d", addr,
				level, err);
		}*/
}
#endif

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
#ifdef CONFIG_BT_NUS_SECURITY_ENABLED
	.security_changed = security_changed,
#endif
};

#if defined(CONFIG_BT_NUS_SECURITY_ENABLED)
static void auth_passkey_display(struct bt_conn *conn, unsigned int passkey)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	// printk("Passkey for %s: %06u", addr, passkey);
}

static void auth_passkey_confirm(struct bt_conn *conn, unsigned int passkey)
{
	char addr[BT_ADDR_LE_STR_LEN];

	auth_conn = bt_conn_ref(conn);

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	// printk("Passkey for %s: %06u", addr, passkey);
	// printk("Press Button 1 to confirm, Button 2 to reject.");
}

static void auth_cancel(struct bt_conn *conn)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	// printk("Pairing cancelled: %s", addr);
}

static void pairing_complete(struct bt_conn *conn, bool bonded)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	// printk("Pairing completed: %s, bonded: %d", addr, bonded);
}

static void pairing_failed(struct bt_conn *conn, enum bt_security_err reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	// printk("Pairing failed conn: %s, reason %d", addr, reason);
}

static struct bt_conn_auth_cb conn_auth_callbacks = {
	.passkey_display = auth_passkey_display,
	.passkey_confirm = auth_passkey_confirm,
	.cancel = auth_cancel,
};

static struct bt_conn_auth_info_cb conn_auth_info_callbacks = {
	.pairing_complete = pairing_complete,
	.pairing_failed = pairing_failed};
#else
static struct bt_conn_auth_cb conn_auth_callbacks;
static struct bt_conn_auth_info_cb conn_auth_info_callbacks;
#endif

static void bt_receive_cb(struct bt_conn *conn, const uint8_t *const data,
						  uint16_t len)
{
#define TEMP_BUFFER_SIZE 100

	char temp_buffer[TEMP_BUFFER_SIZE];
	int err;
	char addr[BT_ADDR_LE_STR_LEN] = {0};

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, ARRAY_SIZE(addr));
#if DEBUG_PRINT
	sprintf(str, "Received data from: %s", addr);
	my_printk(str);
#endif

	for (uint16_t pos = 0; pos != len;)
	{

		/* Keep the last byte of TX buffer for potential LF char. */
		size_t tx_data_size = sizeof(temp_buffer) - 1;
		uint16_t txlen;

		if ((len - pos) > tx_data_size)
		{
			txlen = tx_data_size;
		}
		else
		{
			txlen = (len - pos);
		}

		memcpy(temp_buffer, &data[pos], txlen);

		pos += txlen;

		/* Append the LF character when the CR character triggered
		 * transmission from the peer.
		 */
		if ((pos == len) && (data[len - 1] == '\r'))
		{
			temp_buffer[txlen] = '\n';
			txlen++;
		}
		// add 0
		temp_buffer[txlen] = 0;

		// Parse the received data
#if DEBUG_PRINT
		sprintf(str, "Received over BT: %s\r\n", temp_buffer);
		my_printk(str);
#endif
		err = parse_config(temp_buffer);
		switch (err)
		{
		case SBOX_CONFIG_PARSE_OK:
			bt_nus_send(NULL, "OK\r\n", sizeof("OK\r\n"));
			break;
		case SBOX_CONFIG_PARSE_ERROR_UNKNOWN_PARAM:
			bt_nus_send(NULL, "Error, unknown parameter\r\n", sizeof("Error, unknown parameter\r\n"));
			break;
		case SBOX_CONFIG_PARSE_ERROR_INVALID_ARG:
			bt_nus_send(NULL, "Error, invalid argument\r\n", sizeof("Error, invalid argument\r\n"));
			break;
		}
	}
}

static struct bt_nus_cb nus_cb = {
	.received = bt_receive_cb,
};

#ifdef CONFIG_BT_NUS_SECURITY_ENABLED
static void num_comp_reply(bool accept)
{
	if (accept)
	{
		bt_conn_auth_passkey_confirm(auth_conn);
		// printk("Numeric Match, conn %p", (void *)auth_conn);
	}
	else
	{
		bt_conn_auth_cancel(auth_conn);
		// printk("Numeric Reject, conn %p", (void *)auth_conn);
	}

	bt_conn_unref(auth_conn);
	auth_conn = NULL;
}

void button_changed(uint32_t button_state, uint32_t has_changed)
{
	// uint32_t buttons = button_state & has_changed;

	if (auth_conn)
	{
		num_comp_reply(true);
	}
}
#endif /* CONFIG_BT_NUS_SECURITY_ENABLED */

struct gpio_callback int1_tilt_cb_data;
struct gpio_callback int1_in_cb_data;

void convert_analog_to_message(uint8_t dest[SENSOR_DATA_LEN], int16_t raw, int32_t conv)
{
	uint8_t *byteArray = (uint8_t *)&conv;
	dest[2] = raw >> 8;
	dest[3] = raw & 0xff;
	for (int i = 0; i < 4; ++i)
	{
		dest[4 + i] = byteArray[i];
	}
}

void debug_message(uint8_t *th_name, struct message_item_t *msg)
{
	retained.msg_counter++;
	msg->message[SENSOR_DATA_LEN - 2] = retained.msg_counter >> 8;
	msg->message[SENSOR_DATA_LEN - 1] = retained.msg_counter & 0xff;
#if DEBUG_PRINT
	char buf[10];
	if (retained.debug && !config_in_progress)
	{
		k_sem_take(&debug_message_sem, K_MSEC(2500));
		sprintf(str, "Debug: Thread: %s ", th_name);
		my_printk(str);
		sprintf(str, "");
		for (uint8_t i = 0; i < SENSOR_DATA_LEN; i++)
		{
			sprintf(buf, "%02x ", msg->message[i]);
			strcat(str, buf);
		}
		strcat(str, "\r\n");
		my_printk(str);
		k_sem_give(&debug_message_sem);
	}
#endif
}
/*
void thread_config(void *, void *, void *) {
	struct message_item_t *message;
	int i =0;
//	uint32_t compile_timestamp = BUILD_TIMESTAMP;
	uint32_t ts;
#if DEBUG_PRINT
	if (retained.debug) {
		my_printk("Starting config thread\r\n");
	}
#endif
	while(1) {
		message = k_malloc(message_item_size);
		memset(message,0,sizeof(struct message_item_t));
		if(i%2 == 0) {
			message->message[0]=255;
			message->message[1]=0;
			message->message[2]=retained.timeout >> 8;
			message->message[3]=retained.timeout & 0xff;
			//message->message[4]=retained.sleep_loop >> 8;
			//message->message[5]=retained.sleep_loop & 0xff;
			message->message[4]=retained.wakeup_source >> 8;
			message->message[5]=retained.wakeup_source & 0xff;
			message->message[6]=retained.sensor_select >> 8;
			message->message[7]=retained.sensor_select & 0xff;
			ts = retained.config_timestamp;
			memcpy(&message->message[8], &ts,4);

		} else {
			message->message[0]=254;
			message->message[1]=0;
			memcpy(&message->message[2],&calibration_data.uintArray[5],4);
			memcpy(&message->message[6],&calibration_data.uintArray[6],4);
			memcpy(&message->message[10],&compile_timestamp,4);
		}
		i++;
		debug_message("config", message);
		k_sem_take(&message_fifo_sem, K_FOREVER);
		k_fifo_put(&message_fifo,message);
		k_sem_give(&message_fifo_sem);
		k_msleep(60*1000);
	}
}
*/
void thread_accel(void *, void *, void *)
{
	struct message_item_t *message;
	inv_imu_sensor_event_t imu_event;
	int16_t temperature;
	uint8_t i = 0;
	int ret;
	while(!main_ready) {
		k_msleep(100);
	}
	while (1)
	{
		if (retained.sensor_time_config[SBOX_CONFIG_SENSOR_ACCEL] > 0)
		{
			if (!accel_initialized)
			{
#if DEBUG_PRINT
				my_printk("Starting accel/gyro thread\r\n");
#endif
				gyro_configureWOM(gyro_irq_handler, retained.accel_sensitivity, retained.accel_sensitivity, retained.accel_sensitivity, retained.accel_freq, true);
				gyro_startAccel(50, 2);
				// gyro_startGyro(50, 250);
				accel_initialized = 1;
				k_msleep(100);
			}
			k_sem_take(&gyro_sem, K_FOREVER);
			ret = gyro_getDataFromRegisters(&imu_event);
			k_sem_give(&gyro_sem);
			message = k_malloc(message_item_size);
			if(message == NULL) {
				malloc_error = 1;
			}
			memset(message, 0, sizeof(struct message_item_t));
			if (i % 2)
			{
				message->message[0] = SBOX_CONFIG_SENSOR_ACCEL;
				message->message[1] = 0;
				message->message[2] = imu_event.accel[0] >> 8;
				message->message[3] = imu_event.accel[0] & 0xff;
				message->message[4] = imu_event.accel[1] >> 8;
				message->message[5] = imu_event.accel[1] & 0xff;
				message->message[6] = imu_event.accel[2] >> 8;
				message->message[7] = imu_event.accel[2] & 0xff;
				temperature = ((100 * imu_event.temperature) / 128) + 2500;
				message->message[8] = temperature >> 8;
				message->message[9] = temperature & 0xff;
				debug_message("accel", message);
			}
			else
			{
				message->message[0] = SBOX_CONFIG_SENSOR_GYRO;
				message->message[1] = 0;
				message->message[2] = imu_event.gyro[0] >> 8;
				message->message[3] = imu_event.gyro[0] & 0xff;
				message->message[4] = imu_event.gyro[1] >> 8;
				message->message[5] = imu_event.gyro[1] & 0xff;
				message->message[6] = imu_event.gyro[2] >> 8;
				message->message[7] = imu_event.gyro[2] & 0xff;
				temperature = ((100 * imu_event.temperature) / 128) + 2500;
				message->message[8] = temperature >> 8;
				message->message[9] = temperature & 0xff;
				debug_message("gyro", message);
			}
			i++;
			k_sem_take(&message_fifo_sem, K_FOREVER);
			k_fifo_put(&message_fifo, message);
			k_sem_give(&message_fifo_sem);
			k_msleep(1000 * retained.sensor_time_config[SBOX_CONFIG_SENSOR_ACCEL]);
		}
		else
		{
			k_msleep(1000);
		}
	}
}

void thread_ping(void *, void *, void *)
{
	struct message_item_t *message;
	uint8_t *byteArray = (uint8_t *)&serial_number;
	while(!main_ready) {
		k_msleep(100);
	}
#if DEBUG_PRINT
	my_printk("Starting ping thread\r\n");
#endif
	int16_t analog_raw;
	int32_t analog_mV;
	uint8_t i;	
	while (1)
	{
		analog_raw = AnalogReadRaw(ANALOG_INPUT_DC_IN);
		analog_mV = AnalogConvert_mV(ANALOG_INPUT_DC_IN, analog_raw);
		message = k_malloc(message_item_size);
		if(message == NULL) {
			malloc_error = 1;
		}
		memset(message, 0, sizeof(struct message_item_t));
		if (analog_mV >= 7000)
		{
			message->message[0] = SBOX_CONFIG_SENSOR_DC_IN;
			message->message[1] = 0;
		}
		else
		{
			analog_raw = AnalogReadRaw(ANALOG_INPUT_BATTERY);
			analog_mV = AnalogConvert_mV(ANALOG_INPUT_BATTERY, analog_raw);
			message->message[0] = SBOX_CONFIG_SENSOR_BATTERY;
			message->message[1] = 0;
		}
		convert_analog_to_message(message->message, analog_raw, analog_mV);
		for (i = 0; i < 4; ++i)
		{
			message->message[8 + i] = byteArray[i];
		}
		debug_message("ping", message);
		k_sem_take(&message_fifo_sem, K_FOREVER);
		k_fifo_put(&message_fifo, message);
		k_sem_give(&message_fifo_sem);
		k_msleep(1000);
	}
}

void thread_analog(void *, void *, void *)
{
	struct message_item_t *message;
	int16_t analog_raw=0;
	int32_t analog_mV=0;
	while(!main_ready) {
		k_msleep(100);
	}
#if DEBUG_PRINT
	if (retained.debug)
	{
		my_printk("Starting analog thread\r\n");
	}
#endif
	uint8_t i = 0;	
	while (1)
	{
		if (retained.sensor_time_config[SBOX_CONFIG_SENSOR_ANALOG_IN] > 0)
		{
			message = k_malloc(message_item_size);
			if(message == NULL) {
				malloc_error = 1;
			}
			memset(message->message, 0, message_item_size);
			switch (i % 2)
			{
			case 0:
				analog_raw = AnalogReadRaw(ANALOG_INPUT_1);
				analog_mV = AnalogConvert_mV(ANALOG_INPUT_1, analog_raw);
				message->message[0] = SBOX_CONFIG_SENSOR_ANALOG_IN;
				message->message[1] = 0;
				break;
			case 1:
				analog_raw = AnalogReadRaw(ANALOG_INPUT_20mA);
				analog_mV = AnalogConvert_mV(ANALOG_INPUT_20mA, analog_raw);
				message->message[0] = SBOX_CONFIG_SENSOR_20MA;
				message->message[1] = 0;
				break;
			}
			i++;
			convert_analog_to_message(message->message, analog_raw, analog_mV);
			debug_message("analog", message);
			k_sem_take(&message_fifo_sem, K_FOREVER);
			k_fifo_put(&message_fifo, message);
			k_sem_give(&message_fifo_sem);
			k_msleep(1000 * retained.sensor_time_config[SBOX_CONFIG_SENSOR_ANALOG_IN]);
		}
		else
		{
			k_msleep(1000);
		}
	}
}

void thread_counter(void *, void *, void *)
{
	struct message_item_t *message;
	while(!main_ready) {
		k_msleep(100);
	}
#if DEBUG_PRINT
	if (retained.debug)
	{
		my_printk("Starting counter thread\r\n");
	}
#endif
	uint8_t i = 0;	
	while (1)
	{
		if (retained.sensor_time_config[SBOX_CONFIG_SENSOR_TILT_COUNTER] > 0)
		{
			if (!rtc_initialized)
			{
				counter_timers_init();
			}
			message = k_malloc(message_item_size);
			if(message == NULL) {
				malloc_error = 1;
			}
			memset(message->message, 0, message_item_size);
			message->message[0] = SBOX_CONFIG_SENSOR_TILT_COUNTER + (i % 2);
			message->message[1] = 0;
			if (i % 2 == 0)
			{
				memcpy(&message->message[2], (const void *)&tilt_counter, 4);
				memcpy(&message->message[6], (const void *)&total_tilt_counter, 4);
			}
			else
			{
				memcpy(&message->message[2], (const void *)&in_counter, 4);
				memcpy(&message->message[6], (const void *)&total_in_counter, 4);
			}
			i++;
			debug_message("counter", message);
			k_sem_take(&message_fifo_sem, K_FOREVER);
			k_fifo_put(&message_fifo, message);
			k_sem_give(&message_fifo_sem);
			k_msleep(1000 * retained.sensor_time_config[SBOX_CONFIG_SENSOR_TILT_COUNTER]);
		}
		else
		{
			k_msleep(1000);
		}
	}
}


void thread_rs232(void *, void *, void *) {
	struct message_item_t *message;
	//char buf[UART_RX_BUFFER_SIZE+2];
	while(!main_ready) {
		k_msleep(100);
	}
	while(1) {
		if(retained.sensor_time_config[SBOX_CONFIG_SENSOR_RS232]>0) {
			if(!rs232_initialized) {
				uarte_init(UARTE_MODE_RS232,retained.uart_baud);
				if (retained.debug) {
					my_printk("Starting rs232 thread\r\n");
				}
				rs485_initialized = 0;
				rs232_initialized = 1;
			}
			message = k_malloc(message_item_size);
			if(message == NULL) {
				malloc_error = 1;
			}
			memset(message,0,sizeof(struct message_item_t));
			message->message[0]=SBOX_CONFIG_SENSOR_RS232;
			//if(nrfx_uarte_rx_new_data_check(&uarte_instance) || uarte_new_data) {
			read_uart_input(str);
			my_printk("\r\n");
			if(sscanf(str,"%f",&rs_readings_union.float_val)==1){

				sprintf(str,"Readed: %f\r\n",rs_readings_union.float_val);
				my_printk(str);

				memcpy(&(message->message[2]),&(rs_readings_union.uint_val),4); // sprawdzić wskaźniki - stara wartość memcpy(message->message[2],&rs_readings_union.uint_val,4);
				debug_message("rs232", message);
				k_sem_take(&message_fifo_sem, K_FOREVER);
				k_fifo_put(&message_fifo,message);
				k_sem_give(&message_fifo_sem);
			}
			
			/*
			if(nrfx_uarte_rx_new_data_check(&uarte_instance) || uarte_new_data) {
				if(k_sem_take(&uarte_rx_sem, K_MSEC(100))!=0) {
					nrfx_uarte_rx_abort(&uarte_instance, false, false);
					k_msleep(1);
					k_sem_take(&uarte_rx_sem, K_MSEC(10));

				}
				if(rx_in_progress == 0 && uarte_new_data) {
					message->message[1]=uarte_new_data;
					memcpy(buf,uarte_rx_buffer,uarte_new_data);
					buf[uarte_new_data]=0;
					memcpy(message->message+2,buf,uarte_new_data);
					uarte_new_data=0;
					sprintf(str,"RS232 data len: %d, data: %s\r\n", message->message[1], buf);
					my_printk(str);

				}
				nrfx_uarte_rx(&uarte_instance, uarte_rx_buffer,sizeof(uarte_rx_buffer));
				rx_in_progress = 1;
			}
			*/
			
			k_msleep(1000 * retained.sensor_time_config[SBOX_CONFIG_SENSOR_RS232]);
		} else {
			k_msleep(1000);
		}
	}
}
/*
void thread_rs485(void *, void *, void *) {
	struct message_item_t *message;
	char buf[UART_RX_BUFFER_SIZE+2];
	while(!main_ready) {
		k_msleep(100);
	}
	while(1) {
		if(retained.sensor_time_config[SBOX_CONFIG_SENSOR_RS485]>0) {
			if(!rs485_initialized) {
				uarte_init(UARTE_MODE_RS485);
				if (retained.debug) {
					my_printk("Starting rs485 thread\r\n");
				}
				rs485_initialized = 1;
				rs232_initialized = 0;
			}
			message = k_malloc(message_item_size);
			memset(message,0,sizeof(struct message_item_t));
			message->message[0]=SBOX_CONFIG_SENSOR_RS485;

			if(nrfx_uarte_rx_new_data_check(&uarte_instance) || uarte_new_data) {
				if(k_sem_take(&uarte_rx_sem, K_MSEC(100))!=0) {
					nrfx_uarte_rx_abort(&uarte_instance, false, false);
					k_msleep(1);
					k_sem_take(&uarte_rx_sem, K_MSEC(10));

				}
				if(rx_in_progress == 0 && uarte_new_data) {
					message->message[1]=uarte_new_data;
					memcpy(buf,uarte_rx_buffer,uarte_new_data);
					buf[uarte_new_data]=0;
					memcpy(message->message+2,buf,uarte_new_data);
					uarte_new_data=0;
					sprintf(str,"RS485 data len: %d, data: %s\r\n", message->message[1], buf);
					my_printk(str);

				}
				nrfx_uarte_rx(&uarte_instance, uarte_rx_buffer,sizeof(uarte_rx_buffer));
				rx_in_progress = 1;
			}

			debug_message("rs485", message);
			k_sem_take(&message_fifo_sem, K_FOREVER);
			k_fifo_put(&message_fifo,message);
			k_sem_give(&message_fifo_sem);
			k_msleep(1000 * retained.sensor_time_config[SBOX_CONFIG_SENSOR_RS485]);
		} else {
			k_msleep(1000);
		}
	}
}
*/
/*
void thread_1wire(void *, void *, void *) {
	struct message_item_t *message;
	if (retained.debug) {
		my_printk("Starting 1wire thread\r\n");
	}
	while(!main_ready) {
		k_msleep(100);
	}
	while(1) {
		if(retained.sensor_time_config[SBOX_CONFIG_SENSOR_1WIRE]>0) {
			message = k_malloc(message_item_size);
			message->message[0]=SBOX_CONFIG_SENSOR_1WIRE;
			message->message[1]=0;

			debug_message("1wire", message);
			k_sem_take(&message_fifo_sem, K_FOREVER);
			k_fifo_put(&message_fifo,message);
			k_sem_give(&message_fifo_sem);
			k_msleep(1000 * retained.sensor_time_config[SBOX_CONFIG_SENSOR_1WIRE]);
		} else {
			k_msleep(1000);
		}
	}
}
*/
void thread_message(void *, void *, void *)
{
	struct message_item_t *message;
	int err;
	while(!main_ready) {
		k_msleep(20);
	}
#if DEBUG_PRINT
	if (retained.debug)
	{
		my_printk("Starting message thread\r\n");
	}
#endif	
	while (1)
	{
		k_sem_take(&message_fifo_sem, K_FOREVER);
		message = k_fifo_get(&message_fifo, K_NO_WAIT);
		k_sem_give(&message_fifo_sem);
		if (message != NULL)
		{
			for (uint8_t i = 0; i < SENSOR_DATA_LEN; i++)
			{
				sensor_data[i] = message->message[i];
			}
			if (!config_in_progress)
			{
				err = bt_le_adv_update_data(ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
#if DEBUG_PRINT
				if (err)
				{
					sprintf(str, "Advertising failed to update (err %d)\r\n", err);
					my_printk(str);
				}
#endif
			}
			k_free(message);
		}
		k_msleep(350);
	}
}

void thread_callback(struct thread_analyzer_info *info)
{
	sprintf(str, "Th analyzer: %s ", info->name),
		my_printk(str);
	k_msleep(2);
	sprintf(str, "size: %d used %d\r\n", info->stack_size, info->stack_used);
	my_printk(str);
	k_msleep(2);
}

bool sensor_is_enabled(uint8_t id)
{
	uint16_t mask = 1 << id;
	return (retained.sensor_select & mask) != 0;
}

bool wakeup_is_enabled(uint8_t id)
{
	uint16_t mask = 1 << id;
	return (retained.wakeup_source & mask) != 0;
}


K_THREAD_DEFINE(th_ping, SENSOR_STACKSIZE, thread_ping, NULL, NULL, NULL, SENSOR_THREAD_PRIORITY, 0, SENSOR_THREAD_STARTUP_DELAY);
K_THREAD_DEFINE(th_accel, SENSOR_STACKSIZE, thread_accel, NULL, NULL, NULL, SENSOR_THREAD_PRIORITY, 0, SENSOR_THREAD_STARTUP_DELAY);
K_THREAD_DEFINE(th_analog, SENSOR_STACKSIZE, thread_analog, NULL, NULL, NULL, SENSOR_THREAD_PRIORITY, 0, SENSOR_THREAD_STARTUP_DELAY);
K_THREAD_DEFINE(th_counter, SENSOR_STACKSIZE, thread_counter, NULL, NULL, NULL, SENSOR_THREAD_PRIORITY, 0, SENSOR_THREAD_STARTUP_DELAY);
K_THREAD_DEFINE(th_rs232, SENSOR_STACKSIZE, thread_rs232, NULL, NULL, NULL, SENSOR_THREAD_PRIORITY, 0, SENSOR_THREAD_STARTUP_DELAY);

// K_THREAD_DEFINE(th_rs485, SENSOR_STACKSIZE, thread_rs485, NULL, NULL, NULL, SENSOR_THREAD_PRIORITY, 0, SENSOR_THREAD_STARTUP_DELAY);
// K_THREAD_DEFINE(th_1wire, SENSOR_STACKSIZE, thread_1wire, NULL, NULL, NULL, SENSOR_THREAD_PRIORITY, 0, SENSOR_THREAD_STARTUP_DELAY);
K_THREAD_DEFINE(th_message, SENSOR_STACKSIZE, thread_message, NULL, NULL, NULL, MESSAGE_THREAD_PRIORITY, 0, SENSOR_THREAD_STARTUP_DELAY);

void suspend_threads(void)
{
	// k_thread_suspend(th_config); // newer suspend
	// k_thread_suspend(th_ping);   // newer suspend

	k_thread_suspend(th_accel);
	k_thread_suspend(th_analog);
	k_thread_suspend(th_counter);
	k_thread_suspend(th_rs232);
	// k_thread_suspend(th_rs485);
	// k_thread_suspend(th_1wire);
	// k_thread_suspend(th_message); // newer suspend
}

void resume_configured_threads(void)
{
	gyro_reinit();
	accel_initialized = 0;
	gyro_initialized = 0;
	rs232_initialized = 0;
	rs485_initialized = 0;
	if (sensor_is_enabled(SBOX_CONFIG_SENSOR_ACCEL))
	{
		k_thread_resume(th_accel);
	}
	if (sensor_is_enabled(SBOX_CONFIG_SENSOR_ANALOG_IN))
	{
		k_thread_resume(th_analog);
	}
	if (sensor_is_enabled(SBOX_CONFIG_SENSOR_TILT_COUNTER))
	{
		k_thread_resume(th_counter);
	}
	
	if(sensor_is_enabled(SBOX_CONFIG_SENSOR_RS232)) {
		k_thread_resume(th_rs232);
	}
	/*	if(sensor_is_enabled(SBOX_CONFIG_SENSOR_RS485)) {
			k_thread_resume(th_rs485);
		}
		*/
	/*
	if(sensor_is_enabled(SBOX_CONFIG_SENSOR_1WIRE)) {
		k_thread_resume(th_1wire);
	}*/
}

void configure_lpcomp(void)
{
#if (KERNEL_VERSION_MAJOR == 3) && (KERNEL_VERSION_MINOR >= 6)	
	switch (retained.lpcomp_setup & SBOX_CONFIG_LPCOMP_DETECT_MASK)
	{
	case 0:
		lpcomp_config.detection = NRF_LPCOMP_DETECT_CROSS;
		break;
	case 0x10:
		lpcomp_config.detection = NRF_LPCOMP_DETECT_UP;
		break;
	case 0x20:
		lpcomp_config.detection = NRF_LPCOMP_DETECT_DOWN;
		break;
	case 0x30:
#if DEBUG_PRINT
		my_printk("Invalid LPCOMP DETECT configuration\r\n");
#endif
	}
	if (retained.lpcomp_setup & SBOX_CONFIG_LPCOMP_HYST_MASK)
	{
		lpcomp_config.hyst = NRF_LPCOMP_HYST_ENABLED;
	}
	else
	{
		lpcomp_config.hyst = NRF_LPCOMP_HYST_NOHYST;
	}
	switch (retained.lpcomp_setup & SBOX_CONFIG_LPCOMP_REF_MASK)
	{
	case 0:
#if DEBUG_PRINT
		my_printk("Invalid LPCOMP REF configuration\r\n");
#endif
		break;
	case 1:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_1_16;
		break;
	case 2:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_1_8;
		break;
	case 3:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_3_16;
		break;
	case 4:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_2_8;
		break;
	case 5:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_5_16;
		break;
	case 6:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_3_8;
		break;
	case 7:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_7_16;
		break;
	case 8:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_4_8;
		break;
	case 9:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_9_16;
		break;
	case 10:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_5_8;
		break;
	case 11:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_11_16;
		break;
	case 12:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_6_8;
		break;
	case 13:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_13_16;
		break;
	case 14:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_7_8;
		break;
	case 15:
		lpcomp_config.reference = NRF_LPCOMP_REF_SUPPLY_15_16;
		break;
#else
	switch(retained.lpcomp_setup & SBOX_CONFIG_LPCOMP_DETECT_MASK) {
		case 0:
			lpcomp_config.config.detection=NRF_LPCOMP_DETECT_CROSS;
			break;
		case 0x10:
			lpcomp_config.config.detection=NRF_LPCOMP_DETECT_UP;
			break;
		case 0x20:
			lpcomp_config.config.detection=NRF_LPCOMP_DETECT_DOWN;
			break;
		case 0x30:
#if DEBUG_PRINT		
			my_printk("Invalid LPCOMP DETECT configuration\r\n");
#endif			
	}
	if(retained.lpcomp_setup & SBOX_CONFIG_LPCOMP_HYST_MASK) {
		lpcomp_config.config.hyst=NRF_LPCOMP_HYST_ENABLED;
	} else {
		lpcomp_config.config.hyst=NRF_LPCOMP_HYST_NOHYST;
	}
	switch(retained.lpcomp_setup & SBOX_CONFIG_LPCOMP_REF_MASK) {
		case 0:
#if DEBUG_PRINT		
			my_printk("Invalid LPCOMP REF configuration\r\n");
#endif			
			break;
		case 1:	
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_1_16;
			break;
		case 2:
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_1_8;	
			break;
		case 3:	
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_3_16;
			break;
		case 4:
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_2_8;	
			break;
		case 5:	
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_5_16;
			break;
		case 6:
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_3_8;	
			break;	
		case 7:	
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_7_16;
			break;
		case 8:
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_4_8;	
			break;	
		case 9:	
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_9_16;
			break;
		case 10:
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_5_8;	
			break;	
		case 11:	
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_11_16;
			break;
		case 12:
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_6_8;	
			break;
		case 13:	
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_13_16;
			break;
		case 14:
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_7_8;	
			break;
		case 15:	
			lpcomp_config.config.reference=NRF_LPCOMP_REF_SUPPLY_15_16;
			break;

#endif
	default:
#if DEBUG_PRINT
		my_printk("Invalid LPCOMP REF configuration\r\n");
#endif
	}
}

void configure_wakeup_sources(void)
{
	int ret = 0;
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_ACCEL))
	{
		// gyro_reinit();
		// ret = gyro_configureWOM(gyro_irq_handler,retained.accel_sensitivity,retained.accel_sensitivity,retained.accel_sensitivity, retained.accel_freq, false);

		// k_msleep(200);
#if DEBUG_PRINT
		if (ret != 0)
		{
			my_printk("ICM42670P config failed\r\nd");
		}
		else
		{
			my_printk("ICM42670P reinitialized\r\n");
		}
#endif
#if DEBUG_PRINT
		sprintf(str, "Sensitivity: %d, freq: %d\r\n", retained.accel_sensitivity, retained.accel_freq);
		my_printk(str);
#endif

#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup accel enabled\r\n");
		}
#endif
		k_msleep(250);
		// int pin1 = DT_GPIO_PIN(DT_NODELABEL(gyro_int1), gpios);
		// int pin2 = DT_GPIO_PIN(DT_NODELABEL(gyro_int2), gpios);
		// nrf_gpio_cfg_sense_input(pin1,NRF_GPIO_PIN_NOPULL, NRF_GPIO_PIN_SENSE_LOW);
		// nrf_gpio_cfg_sense_input(pin2,NRF_GPIO_PIN_NOPULL, NRF_GPIO_PIN_SENSE_LOW);
		// nrf_gpio_cfg_sense_input(pin1,NRF_GPIO_PIN_PULLUP, NRF_GPIO_PIN_SENSE_LOW);
		// nrf_gpio_cfg_sense_input(pin2,NRF_GPIO_PIN_PULLUP, NRF_GPIO_PIN_SENSE_LOW);
		// NRF_GPIOTE->EVENTS_PORT |= (1 << pin1);
		// NRF_GPIOTE->EVENTS_PORT |= (1 << pin1);
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_TILT_COUNTER))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup tilt_counter enabled\r\n");
		}
#endif
		nrf_gpio_cfg_input(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(tilt_int1), gpios), NRF_GPIO_PIN_NOPULL);
		nrf_gpio_cfg_sense_set(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(tilt_int1), gpios), NRF_GPIO_PIN_SENSE_LOW);
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_IN_COUNTER))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup in_counter enabled\r\n");
		}
#endif
		nrf_gpio_cfg_input(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(in_int1), gpios), NRF_GPIO_PIN_NOPULL);
		nrf_gpio_cfg_sense_set(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(in_int1), gpios), NRF_GPIO_PIN_SENSE_LOW);
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_ONOFF))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup on/off enabled\r\n");
		}
#endif
		nrf_gpio_cfg_input(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(onoff_int1), gpios), NRF_GPIO_PIN_NOPULL);
		nrf_gpio_cfg_sense_set(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(onoff_int1), gpios), NRF_GPIO_PIN_SENSE_LOW);
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_RS232))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup rs232 enabled\r\n");
		}
#endif
		nrf_gpio_cfg_input(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(rs232_int1), gpios), NRF_GPIO_PIN_NOPULL);
		nrf_gpio_cfg_sense_set(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(rs232_int1), gpios), NRF_GPIO_PIN_SENSE_LOW);
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_RS485))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup rs485 enabled\r\n");
		}
#endif
		nrf_gpio_cfg_input(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(rs485_int1), gpios), NRF_GPIO_PIN_NOPULL);
		nrf_gpio_cfg_sense_set(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(rs485_int1), gpios), NRF_GPIO_PIN_SENSE_LOW);
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_NFC))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup nfc enabled (TODO)\r\n");
		}
#endif
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_1WIRE))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup 1wire enabled\r\n");
		}
#endif
		nrf_gpio_cfg_input(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(wire_int1), gpios), NRF_GPIO_PIN_NOPULL);
		nrf_gpio_cfg_sense_set(NRF_DT_GPIOS_TO_PSEL(DT_NODELABEL(wire_int1), gpios), NRF_GPIO_PIN_SENSE_LOW);
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_ANALOG_IN))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup analog_in enabled\r\n");
		}
#endif
		configure_lpcomp();
		lpcomp_config.input = NRF_LPCOMP_INPUT_0;
		nrfx_err = nrfx_lpcomp_init(&lpcomp_config, comparator_handler);
#if DEBUG_PRINT
		if (nrfx_err != NRFX_SUCCESS)
		{
			sprintf(str, "nrfx_lpcomp_init error: %08x", nrfx_err);
			my_printk(str);
		}
#endif
		nrfx_lpcomp_enable();
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_DC_IN))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup dc_in enabled\r\n");
		}
#endif
		configure_lpcomp();
		lpcomp_config.input = NRF_LPCOMP_INPUT_1;
		nrfx_err = nrfx_lpcomp_init(&lpcomp_config, comparator_handler);
#if DEBUG_PRINT
		if (nrfx_err != NRFX_SUCCESS)
		{
			sprintf(str, "nrfx_lpcomp_init error: %08x", nrfx_err);
			my_printk(str);
		}
#endif
		nrfx_lpcomp_enable();
	}
	if (wakeup_is_enabled(SBOX_CONFIG_WAKEUP_20MA))
	{
#if DEBUG_PRINT
		if (retained.debug)
		{
			my_printk("Wakeup 20mA enabled\r\n");
		}
#endif
		configure_lpcomp();
		lpcomp_config.input = NRF_LPCOMP_INPUT_3;
		nrfx_err = nrfx_lpcomp_init(&lpcomp_config, comparator_handler);
#if DEBUG_PRINT
		if (nrfx_err != NRFX_SUCCESS)
		{
			sprintf(str, "nrfx_lpcomp_init error: %08x", nrfx_err);
			my_printk(str);
		}
#endif
		nrfx_lpcomp_enable();
	}
}

void read_uart_input(char *buf)
{
	bool input_end;
	uint8_t i;
	char buf2[UART_RX_BUFFER_SIZE + 2];
	input_end = false;
	i = 0;
	buf[0] = 0;
	while (!input_end)
	{
#if (KERNEL_VERSION_MAJOR == 3) && (KERNEL_VERSION_MINOR >= 6)		
		if (nrfx_uarte_rx_new_data_check(&uarte_instance) || uarte_new_data)
#else
		if(uarte_new_data)
#endif		
		{
			if (k_sem_take(&uarte_rx_sem, K_MSEC(20)) != 0)
			{
				nrfx_uarte_rx_abort(&uarte_instance, false, false);
				k_msleep(1);
				k_sem_take(&uarte_rx_sem, K_MSEC(5));
			}
			if (uarte_new_data)
			{
				memcpy(buf2, uarte_rx_buffer, uarte_new_data);
				buf2[uarte_new_data] = 0;
				my_printk(buf2);
				if (buf2[0] == '\n' || buf2[0] == '\r' || i >= INPUT_BUFFER_LEN - 1)
				{
					input_end = true;
				}
				else
				{
					i += uarte_new_data;
					uarte_new_data = 0;
					strcat(buf, buf2);
				}
			}
		}
		nrfx_uarte_rx(&uarte_instance, uarte_rx_buffer, 1);		
		k_msleep(1);
	}
}

void read_calibration_data(void)
{
	int *ptr;
	uint8_t i;

	for (i = 0; i < 9; i++)
	{
		ptr = (int *)&NRF_UICR->CUSTOMER[i];
		// sprintf(str,"Reading %p\r\n",(void*)ptr);
		// my_printk(str);
		calibration_data.uintArray[i] = nrfx_nvmc_uicr_word_read(ptr);
	}
	// read MAC
	ptr = (int *)&NRF_FICR->DEVICEADDR[0];
	mac_address[0] = nrfx_nvmc_uicr_word_read(ptr);
	ptr = (int *)&NRF_FICR->DEVICEADDR[1];
	mac_address[1] = nrfx_nvmc_uicr_word_read(ptr);

	sprintf(str, "MAC: %02X:%02X:%02X:%02X:%02X:%02X\r\n",
			((mac_address[1] >> 8) & 0xFF) | 0xC0, // 2 MSB bit must be 11
			mac_address[1] & 0xFF,
			(mac_address[0] >> 24) & 0xFF,
			(mac_address[0] >> 16) & 0xFF,
			(mac_address[0] >> 8) & 0xFF,
			mac_address[0] & 0xFF);
	uarte_send(str, strlen(str));

	if (calibration_data.uintArray[4] == 1)
	{
		sprintf(str, "Calibration data: \033[32mOK\033[0m\r\n");
		uarte_send(str, strlen(str));
		sprintf(str, "Hardware version: %d\r\n", calibration_data.uintArray[5]);
		uarte_send(str, strlen(str));
		sprintf(str, "Timestamp of production: %d\r\n", calibration_data.uintArray[6]);
		uarte_send(str, strlen(str));
		sprintf(str, "Serial number: %d\r\n", calibration_data.uintArray[7]);
		uarte_send(str, strlen(str));
		for (i = 0; i < 4; i++)
		{
			retained.cal_val[i] = calibration_data.floatArray[i];
		}
	}
	else if (calibration_data.uintArray[4] == ~0)
	{
#if DEBUG_PRINT
		my_printk("Calibration data: \033[31mnot programmed !\033[0m\r\n");
#endif
		for (i = 0; i < 4; i++)
		{
			retained.cal_val[i] = 1.0;
			calibration_data.floatArray[i] = 1.0;
		}
	}
	else
	{
#if DEBUG_PRINT
		my_printk("Error !\r\n");
#endif
	}
}

void write_calibration_data(void)
{	
	calibration_data.uintArray[4] = 1;
	while (!nrfx_nvmc_write_done_check()) {}
	nrfx_nvmc_words_write((uint32_t)&NRF_UICR->CUSTOMER, &calibration_data.uintArray,10);
}

void calibrate_analog_inputs(void)
{
	/* Divided prompt because of buf size limit*/
	const char *prompt1 =
		"Calibration:\r\n"
		"e - erase all UICR\r\n"
		"r - read settings\r\n"
		"w - write settings\r\n";
	const char *prompt2 =
		"1,2,3,4 - calibrate input from reference\r\n"
		"h - set hardware version\r\n"
		"d - set date of production\r\n"
		"s - set cal_val=1.0\r\n";
	const char *prompt3 =
		"n - set serial number\r\n"
		"p - print current settings\r\n"
		"t - test current settings\r\n"
		"q - quit calibration\r\n";
	bool quit = false;
	char buf[UART_RX_BUFFER_SIZE + 2];
	bool new_command;
	uint32_t hw_version, timestamp;
	int16_t raw;
	int32_t input_mv;
	while (!quit)
	{
		nrfx_uarte_rx_abort(&uarte_instance, false, false);
		nrfx_uarte_rx(&uarte_instance, uarte_rx_buffer, 1);		
		new_command = false;
		sprintf(str, "%s", prompt1);
		uarte_send(str, strlen(str));
		k_msleep(5);
		sprintf(str, "%s", prompt2);
		uarte_send(str, strlen(str));
		k_msleep(5);
		sprintf(str, "%s", prompt3);
		uarte_send(str, strlen(str));
		k_msleep(5);
		while (!new_command)
		{
#if (KERNEL_VERSION_MAJOR == 3) && (KERNEL_VERSION_MINOR >= 6)			
			if (nrfx_uarte_rx_new_data_check(&uarte_instance) || uarte_new_data)
#else
			if(uarte_new_data)
#endif			
			{
				if (k_sem_take(&uarte_rx_sem, K_MSEC(100)) != 0)
				{
					nrfx_uarte_rx_abort(&uarte_instance, false, false);
					k_msleep(1);
					k_sem_take(&uarte_rx_sem, K_MSEC(10));
				}
				if (uarte_new_data)
				{
					memcpy(buf, uarte_rx_buffer, uarte_new_data);
					buf[uarte_new_data] = 0;
					uarte_new_data = 0;
					new_command = true;
				}
				nrfx_uarte_rx(&uarte_instance, uarte_rx_buffer, 1);				
			}
			k_msleep(100);
		}
		switch (buf[0])
		{
		case 'e':
			sprintf(str, "Erasing UICR... ");
			uarte_send(str, strlen(str));
			if (NRFX_SUCCESS == nrfx_nvmc_uicr_erase())
			{
				sprintf(str, "OK\r\n");
				uarte_send(str, strlen(str));
			}
			else
			{
				sprintf(str, "Error !\r\n");
				uarte_send(str, strlen(str));
			}
			break;
		case 'r':
			sprintf(str, "Reading settings ...");
			uarte_send(str, strlen(str));
			read_calibration_data();
			break;
		case 'w':
			sprintf(str, "Writing settings ...");
			uarte_send(str, strlen(str));
			write_calibration_data();
			sprintf(str, "OK\r\n");
			uarte_send(str, strlen(str));
			break;
		case 's':
			uint8_t i;
			sprintf(str, "Set cal_val=1.0");
			uarte_send(str, strlen(str));
			for (i = 0; i < 4; i++)
			{
				retained.cal_val[i] = 1.0;
				calibration_data.floatArray[i] = 1.0;
			}
		case '1':
		case '2':
		case '3':
		case '4':
			sprintf(str, "Calibrating input %c: ", buf[0]);
			uarte_send(str, strlen(str));
			retained.cal_val[buf[0] - '1'] = 1.0;
			k_msleep(200);
			switch (buf[0])
			{
			case '1':
				raw = AnalogReadRaw(ANALOG_INPUT_1);
				input_mv = AnalogConvert_mV(ANALOG_INPUT_1, raw);
				retained.cal_val[0] = 12000.0f / (float)input_mv;
				calibration_data.floatArray[0] = retained.cal_val[0];
				break;
			case '2':
				raw = AnalogReadRaw(ANALOG_INPUT_DC_IN);
				input_mv = AnalogConvert_mV(ANALOG_INPUT_DC_IN, raw);
				retained.cal_val[1] = 12000.0f / (float)input_mv;
				calibration_data.floatArray[1] = retained.cal_val[1];
			case '3':
				raw = AnalogReadRaw(ANALOG_INPUT_BATTERY);
				input_mv = AnalogConvert_mV(ANALOG_INPUT_BATTERY, raw);
				retained.cal_val[2] = 3300.0f / (float)input_mv;
				calibration_data.floatArray[2] = retained.cal_val[2];
			case '4':
				raw = AnalogReadRaw(ANALOG_INPUT_20mA);
				input_mv = AnalogConvert_mV(ANALOG_INPUT_20mA, raw);
				retained.cal_val[3] = 20950.0f / (float)input_mv;
				calibration_data.floatArray[3] = retained.cal_val[3];
			}
			break;
		case 'p':
			sprintf(str, "Current settings: \r\n");
			uarte_send(str, strlen(str));

			sprintf(str, "Hardware version: %d\r\n", calibration_data.uintArray[5]);
			uarte_send(str, strlen(str));
			sprintf(str, "Production timestamp: %d\r\n", calibration_data.uintArray[6]);
			uarte_send(str, strlen(str));
			sprintf(str, "Serial number: %d\r\n", calibration_data.uintArray[7]);
			uarte_send(str, strlen(str));
			break;
		case 't':
			raw = AnalogReadRaw(ANALOG_INPUT_1);
			input_mv = AnalogConvert_mV(ANALOG_INPUT_1, raw);
			sprintf(str, "Input 1 raw: %d, %d mV\r\n", raw, input_mv);
			uarte_send(str, strlen(str));
			raw = AnalogReadRaw(ANALOG_INPUT_DC_IN);
			input_mv = AnalogConvert_mV(ANALOG_INPUT_DC_IN, raw);
			sprintf(str, "Input DC_IN raw: %d, %d mV\r\n", raw, input_mv);
			uarte_send(str, strlen(str));
			raw = AnalogReadRaw(ANALOG_INPUT_BATTERY);
			input_mv = AnalogConvert_mV(ANALOG_INPUT_BATTERY, raw);
			sprintf(str, "Input battery raw: %d, %d mV\r\n", raw, input_mv);
			uarte_send(str, strlen(str));
			raw = AnalogReadRaw(ANALOG_INPUT_20mA);
			input_mv = AnalogConvert_mV(ANALOG_INPUT_20mA, raw);
			sprintf(str, "Input 20mA raw: %d, %d uA\r\n", raw, input_mv);
			uarte_send(str, strlen(str));
			break;
		case 'h':
			sprintf(str, "Enter hardware version: ");
			uarte_send(str, strlen(str));
			read_uart_input(str);
			if (sscanf(str, "%d", &hw_version) == 1)
			{
				calibration_data.uintArray[5] = hw_version;
				sprintf(str, "\r\nOK\r\n");
				uarte_send(str, strlen(str));
			}
			else
			{
				sprintf(str, "\r\nError\r\n");
				uarte_send(str, strlen(str));
			}
			break;
		case 'd':
			sprintf(str, "Enter production timestamp: ");
			uarte_send(str, strlen(str));
			read_uart_input(str);
			if (sscanf(str, "%d", &timestamp) == 1)
			{
				calibration_data.uintArray[6] = timestamp;
				sprintf(str, "\r\nOK\r\n");
				uarte_send(str, strlen(str));
			}
			else
			{
				sprintf(str, "\r\nError\r\n");
				uarte_send(str, strlen(str));
			}
			break;
		case 'n':
			sprintf(str, "Enter serial number: ");
			uarte_send(str, strlen(str));
			read_uart_input(str);
			if (sscanf(str, "%d", &serial_number) == 1)
			{
				calibration_data.uintArray[7] = serial_number;
				sprintf(str, "\r\nOK\r\n");
				uarte_send(str, strlen(str));
			}
			else
			{
				sprintf(str, "\r\nError\r\n");
				uarte_send(str, strlen(str));
			}
			break;
		case 'q':
			quit = true;
			break;
		default:
			sprintf(str, "Error - unknown command\r\n");
			uarte_send(str, strlen(str));
			break;
		}
	}
}
/*
void timer_handler(struct k_timer *timer)
{
	my_printk("Timer event\r\n");
}
*/
// K_TIMER_DEFINE(event_timer, timer_handler, NULL);

static void my_gpio_callback1(const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{
	// Obsługa zdarzenia niskiego stanu GPIO
	my_printk("Low state detected on GPIO pin int1 !\n");

	// Tutaj można dodać kod do przywrócenia systemu lub wykonać inne operacje
	// przed powrotem do trybu SYSTEM ON.
}
/*
static void my_gpio_callback2(const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{
	// Obsługa zdarzenia niskiego stanu GPIO
	my_printk("Low state detected on GPIO pin int2!\n");

	// Tutaj można dodać kod do przywrócenia systemu lub wykonać inne operacje
	// przed powrotem do trybu SYSTEM ON.
}
*/
struct gpio_callback my_gpio_cb1, my_gpio_cb2;
/*
static void timer_handler(nrf_timer_event_t event_type, void *p_context)
{
	if (event_type == NRF_TIMER_EVENT_COMPARE0)
	{
		tilt_counter = nrfx_timer_capture(&timer1, 0);
		total_tilt_counter += tilt_counter;
		in_counter = nrfx_timer_capture(&timer2, 0);
		total_in_counter += in_counter;
		if (tilt_counter > 0)
		{
			timeout = retained.timeout;
			// my_printk("Tilt counter detected - no sleep\r\n");
		}
		if (in_counter > 0)
		{
			timeout = retained.timeout;
			// my_printk("In counter detected - no sleep\r\n");
		}
		nrfx_timer_clear(&timer1);
		nrfx_timer_clear(&timer2);
	}
}
*/
void rtc_event_handler(nrfx_rtc_int_type_t int_type)
{
	switch (int_type)
	{
	case NRFX_RTC_INT_COMPARE0:
		// printk("RTC compare event occurred\n");
		nrfx_rtc_counter_clear(&rtc);
		nrfx_rtc_cc_set(&rtc, CC_CHANNEL, RTC_TIMEOUT_MS / RTC_PRESCALER_MS, true);
		tilt_counter = nrfx_timer_capture(&timer1, 0);
		total_tilt_counter += tilt_counter;
		in_counter = nrfx_timer_capture(&timer2, 0);
		total_in_counter += in_counter;
		if (tilt_counter > 0)
		{
			timeout = retained.timeout;
			// my_printk("Tilt counter detected - no sleep\r\n");
		}
		if (in_counter > 0)
		{
			timeout = retained.timeout;
			// my_printk("In counter detected - no sleep\r\n");
		}
		nrfx_timer_clear(&timer1);
		nrfx_timer_clear(&timer2);
		break;
	default:
		break;
	}
}

void counter_timers_init(void)
{
	rtc_config.prescaler = NRF_RTC_FREQ_TO_PRESCALER(1000 / RTC_PRESCALER_MS);
	nrfx_rtc_init(&rtc, &rtc_config, rtc_event_handler);

	IRQ_CONNECT(DT_IRQN(DT_NODELABEL(rtc2)),
				DT_IRQ(DT_NODELABEL(rtc2), priority),
				nrfx_isr, nrfx_rtc_2_irq_handler, 0);
	irq_enable(DT_IRQN(DT_NODELABEL(rtc2)));

	nrfx_rtc_tick_disable(&rtc);
	nrfx_rtc_counter_clear(&rtc);
	nrfx_rtc_enable(&rtc);
	nrfx_rtc_cc_set(&rtc, CC_CHANNEL, RTC_TIMEOUT_MS / RTC_PRESCALER_MS, true);

	nrf_gpio_cfg_input(int1_tilt.pin, NRF_GPIO_PIN_NOPULL);
	nrf_gpio_cfg_input(int1_in.pin, NRF_GPIO_PIN_NOPULL);

	counter1_init();
	counter2_init();
	gpiote_init(int1_tilt.pin, int1_in.pin);

	ppi_timer_start_counter_init();
	ppi_gpiote_counter_init();

	rtc_initialized = 1;
}


int main(void)
{

	int err;
	int ret;
	char *heap_mem_prt;
	heap_mem_prt = k_malloc(1024);
	if(heap_mem_prt != NULL) {
		memset(heap_mem_prt, 0, 1024);
		k_free(heap_mem_prt);
	}
	settings_subsys_init();
	const struct device *gpio_dev = device_get_binding(DEVICE_DT_NAME(DT_NODELABEL(gpio0)));

	if (IS_ENABLED(CONFIG_APP_RETENTION))
	{
		bool retained_ok = retained_validate();

		if (!retained_ok)
		{
			retained.debug = 1;
			setConfigBit(&retained.wakeup_source, SBOX_CONFIG_DEFAULT_WAKEUP, 1);
			retained.accel_sensitivity = SBOX_ACCEL_DEFAULT_SENSITIVITY;
			retained.accel_freq = SBOX_ACCEL_DEFAULT_FREQ;
			settings_register(&sensor_conf);
			settings_load();
		}

		retained.boots += 1;
		// printk("Retained data: %s\r\n", retained_ok ? "valid" : "INVALID");
		// printk("Boot count: %u\r\n", retained.boots);
		// printk("Off count: %u\r\n", retained.off_count);
		// printk("Active Ticks: %" PRIu64 "\r\n", retained.uptime_sum);
		if (retained.timeout == 0)
		{
			retained.timeout = SBOX_CONFIG_DEFAULT_TIMEOUT;
		}
		if (retained.sleep_loop == 0)
		{
			retained.sleep_loop = SBOX_CONFIG_DEFAULT_SLEEP;
		}
		retained_update();
	}
	else
	{
		// printk("Retained data not supported\r\n");
	}

	uarte_init(UARTE_MODE_RS232,115200);
	// k_timer_start(&uart_send_timer, K_NO_WAIT, K_MSEC(2));

#if DEBUG_PRINT
	sprintf(str, "\r\n\r\n\033[32mStarting %s: ...\033[0m\r\n", CONFIG_BOARD);
	my_printk(str);
	sprintf(str,"Zephyr version: %s\r\n", KERNEL_VERSION_STRING);
	my_printk(str);
#endif

	uint32_t reset_reason = nrf_power_resetreas_get(NRF_POWER);
	nrf_power_resetreas_clear(NRF_POWER, reset_reason);
#if DEBUG_PRINT
	sprintf(str, "Reset reason: %x\r\n", reset_reason);
	my_printk(str);

	sprintf(str, "Compiled at: %s %s\r\n", __DATE__, __TIME__);
	my_printk(str);
#endif

	
	Analog_init();
#if DEBUG_PRINT
	my_printk("SADC initialized\r\n");
#endif	
	read_calibration_data();
	if (calibration_data.uintArray[4] != 1)
	{
		config_in_progress = 1;
		if (!uarte_initialized)
		{
			uarte_init(UARTE_MODE_RS232,retained.uart_baud);
		}
		calibrate_analog_inputs();
		config_in_progress = 0;
	}
	
	if (IS_ENABLED(CONFIG_BT_NUS_SECURITY_ENABLED))
	{
		err = bt_conn_auth_cb_register(&conn_auth_callbacks);
		if (err)
		{
#if DEBUG_PRINT
			my_printk("Failed to register authorization callbacks.\r\n");
#endif
			return 0;
		}

		err = bt_conn_auth_info_cb_register(&conn_auth_info_callbacks);
		if (err)
		{
#if DEBUG_PRINT
			my_printk("Failed to register authorization info callbacks.\r\n");
#endif
			return 0;
		}
	}
	//k_msleep(5);
	// Initialize the Bluetooth Subsystem
	err = bt_enable(NULL);
	if (err)
	{
#if DEBUG_PRINT
		sprintf(str, "Bluetooth init failed (err %d)\r\n", err);
		my_printk(str);
#endif
		return 0;
	}
#if DEBUG_PRINT
	my_printk("Bluetooth initialized\r\n");
#endif

	k_sem_give(&ble_init_ok);
	if (IS_ENABLED(CONFIG_SETTINGS))
	{
		settings_load();
#if DEBUG_PRINT
		my_printk("BT settings loaded\r\n");
#endif
	}
	//k_msleep(10);
	err = bt_nus_init(&nus_cb);
	if (err)
	{
#if DEBUG_PRINT
		sprintf(str, "Failed to initialize UART service (err: %d)", err);
		my_printk(str);
#endif
		return 0;
	}

	//ret = gyro_init();

	if (ret != 0)
	{
#if DEBUG_PRINT
		sprintf(str, "ICM42670P initialization failed: %d", ret);
		my_printk(str);
#endif
		return 0;
	}
	if (!device_is_ready(gpio_dev))
	{
		return 0;
	}

	int pin1 = DT_GPIO_PIN(DT_NODELABEL(gyro_int1), gpios);
	// gpio_pin_configure(gpio_dev, pin1, GPIO_INPUT);
	// gpio_pin_interrupt_configure(gpio_dev, pin1, GPIO_INT_LEVEL_LOW);
	//k_msleep(350);
	//ret = gyro_configureWOM(gyro_irq_handler, retained.accel_sensitivity, retained.accel_sensitivity, retained.accel_sensitivity, retained.accel_freq, false);

#if DEBUG_PRINT
	if (ret != 0)
	{
		my_printk("ICM42670P config failed\r\nd");
	}
	else
	{
		my_printk("ICM42670P initialized\r\n");
	}
#endif	
	thread_analyzer_run(thread_callback);
	err = bt_le_adv_start(BT_LE_ADV_CONN, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
	thread_analyzer_run(thread_callback);
	//k_msleep(50);
	if (err)
	{
#if DEBUG_PRINT
		sprintf(str, "Advertising failed to start (err %d)\r\n", err);
		my_printk(str);
#endif
		return 0;
	}
	else
	{
#if DEBUG_PRINT
		my_printk("BT advertising started\r\n");
#endif
	}

	// Detect external power
	int16_t analog_raw;
	int32_t analog_mV;
	
	analog_raw = AnalogReadRaw(ANALOG_INPUT_DC_IN);
	analog_mV = AnalogConvert_mV(ANALOG_INPUT_DC_IN, analog_raw);
	
#if DEBUG_PRINT
	sprintf(str, "Analog DC_IN: %d\r\n", analog_mV);
	my_printk(str);
#endif

	uint8_t external_power = analog_mV > 7000;

	uint32_t i = 0;
	timeout = retained.timeout;

	if (!device_is_ready(flash_dev))
	{
#if DEBUG_PRINT
		my_printk("Flash device not ready\r\n");
#endif
		return 0;
	}
	/*
	nrfx_uarte_tx_abort(&uarte_instance, true);
	nrfx_uarte_rx_abort(&uarte_instance, true, true);
	nrfx_uarte_uninit(&uarte_instance);
	uarte_initialized = 0;
	rs232_initialized = 0;
*/

	uint32_t ts;
	struct message_item_t *message;
	//uint32_t capt;
#if DEBUG_PRINT
		my_printk("\033[32mMain Ready\033[0m\r\n");
#endif	
	main_ready = true;	
	do
	{
#if DEBUG_PRINT
		if (config_in_progress)
		{
			sprintf(str, "Configuration in progress, %d\r\n", i);
			my_printk(str);
			k_sleep(K_MSEC(4000));
		}
		else
		{
			if (external_power)
			{
				my_printk("Always ON\r\n");
			}
			else
			{
				sprintf(str, "Timeout counter %d\r\n", timeout);
				my_printk(str);
				timeout--;
			}
			sprintf(str, "Tilt counter: %d\r\n", tilt_counter);
			my_printk(str);
			sprintf(str, "In counter: %d\r\n", in_counter);
			my_printk(str);
			sprintf(str, "Tick: %d\r\n", i);
			my_printk(str);
		}

		if (irq_received)
		{
			irq_received = 0;
			gyro_getDataFromRegisters(&imu_event);
			sprintf(str, "Accel:  %d, %d, %d ",
					imu_event.accel[0],
					imu_event.accel[1],
					imu_event.accel[2]);
			my_printk(str);
			sprintf(str, "Temp: %d\r\n", ((100 * imu_event.temperature) / 128) + 2500);
			my_printk(str);
		}
		if (retained.debug == 2)
		{
			// k_sem_take(&debug_message_sem, K_MSEC(5000));
			// thread_analyzer_run(thread_callback);
			// k_sem_give(&debug_message_sem);
			retained.debug = 1;
		}
#endif
		/*
				tilt_counter_th=tilt_counter;
				tilt_counter=0;
				in_counter_th=in_counter;
				in_counter=0;
		*/		
		if (i % 60 == 0)
		{
			message = k_malloc(message_item_size);
			if(message == NULL) {
				malloc_error = 1;
			}
			memset(message, 0, sizeof(struct message_item_t));
			message->message[0] = 255;
			message->message[1] = 0;
			message->message[2] = retained.timeout >> 8;
			message->message[3] = retained.timeout & 0xff;
			// message->message[4]=retained.sleep_loop >> 8;
			// message->message[5]=retained.sleep_loop & 0xff;
			message->message[4] = retained.wakeup_source >> 8;
			message->message[5] = retained.wakeup_source & 0xff;
			message->message[6] = retained.sensor_select >> 8;
			message->message[7] = retained.sensor_select & 0xff;
			ts = retained.config_timestamp;
			memcpy(&message->message[8], &ts, 4);
			debug_message("config", message);
			k_sem_take(&message_fifo_sem, K_FOREVER);
			k_fifo_put(&message_fifo, message);
			k_sem_give(&message_fifo_sem);
		}
		if (i % 60 == 1)
		{
			message = k_malloc(message_item_size);
			if(message == NULL) {
				malloc_error = 1;
			}
			memset(message, 0, sizeof(struct message_item_t));
			message->message[0] = 253;
			message->message[1] = 0;
			memcpy(&(message->message[2]), &calibration_data.uintArray[5], 4); // hardware version
			memcpy(&(message->message[6]), &calibration_data.uintArray[6], 4); // production timestamp
			memcpy(&(message->message[10]), &compile_timestamp, 4);
			debug_message("version", message);
			k_sem_take(&message_fifo_sem, K_FOREVER);
			k_fifo_put(&message_fifo, message);
			k_sem_give(&message_fifo_sem);
		}
		if (i % 60 == 2)
		{
			message = k_malloc(message_item_size);
			if(message == NULL) {
				malloc_error = 1;
			}
			memset(message, 0, sizeof(struct message_item_t));
			message->message[0] = 252;
			message->message[1] = 0;
			memcpy(&(message->message[2]), &calibration_data.uintArray[7], 4); // Serial number
			debug_message("serial num", message);
			k_sem_take(&message_fifo_sem, K_FOREVER);
			k_fifo_put(&message_fifo, message);
			k_sem_give(&message_fifo_sem);
		}
		i++;
		k_sleep(K_MSEC(retained.sleep_loop));
	} while ((retained.timeout > 0 && timeout) || (retained.timeout < 0) || external_power);

	suspend_threads();
	k_thread_suspend(th_ping);
#if DEBUG_PRINT
	my_printk("Configuring wakeup sources...\r\n");
#endif
	// send last frame before suspend

	message = k_malloc(message_item_size);
	if(message == NULL) {
		malloc_error = 1;
	}
	memset(message->message, 0, message_item_size);
	message->message[0] = SBOX_SUSPEND_FRAME;
	debug_message("suspend", message);
	k_sem_take(&message_fifo_sem, K_FOREVER);
	k_fifo_put(&message_fifo, message);
	k_sem_give(&message_fifo_sem);

	configure_wakeup_sources();

#if DEBUG_PRINT
	my_printk("Entering system off; shake to restart\r\n");
#endif

	if (IS_ENABLED(CONFIG_APP_RETENTION))
	{
		/* Update the retained state */
		retained.off_count += 1;
		retained_update();
	}
	// k_sleep(K_MSEC(1000));
	// sys_poweroff();

	// int pin2 = DT_GPIO_PIN(DT_NODELABEL(gyro_int2), gpios);

	// k_timer_start(&event_timer, K_SECONDS(60), K_NO_WAIT);

	k_msleep(25);
	ret = gpio_pin_configure(gpio_dev, pin1, GPIO_INPUT);
	// ret = gpio_pin_configure(gpio_dev, pin2, GPIO_INPUT | GPIO_PULL_UP);
	if (ret)
	{
		my_printk("Error configuring GPIO pin!\r\n");
	}
	gpio_init_callback(&my_gpio_cb1, my_gpio_callback1, BIT(pin1));
	gpio_add_callback(gpio_dev, &my_gpio_cb1);
	// gpio_init_callback(&my_gpio_cb2, my_gpio_callback2, BIT(pin2));
	// gpio_add_callback(gpio_dev, &my_gpio_cb2);
	ret = gpio_pin_interrupt_configure(gpio_dev, pin1, GPIO_INT_LEVEL_LOW);
	// ret = gpio_pin_interrupt_configure(gpio_dev, pin2, GPIO_INT_EDGE_TO_INACTIVE);
	if (ret)
	{
		my_printk("Error configuring GPIO interrupt!\r\n");
	}
	thread_analyzer_run(thread_callback);

	k_sleep(K_MSEC(1000));
	//k_thread_suspend(th_uart);
	k_thread_suspend(th_message);
	bt_le_adv_stop();
	bt_disable();
	nrfx_uarte_uninit(&uarte_instance);
/*
	const struct device *const console = DEVICE_DT_GET(DT_CHOSEN(zephyr_console));

	pm_device_action_run(console, PM_DEVICE_ACTION_SUSPEND);
*/	
	const struct device *i2c_dev = DEVICE_DT_GET(DT_NODELABEL(i2c0));
	pm_device_action_run(i2c_dev, PM_DEVICE_ACTION_SUSPEND);

	NRF_UARTE0->TASKS_STOPTX = 1;
	NRF_UARTE0->TASKS_STOPRX = 1;
	NRF_SAADC->TASKS_STOP = 1;
	NRF_TWIM0->TASKS_STOP = 1;
	k_sleep(K_MSEC(100));

	NRF_UARTE0->ENABLE = 0; // disable UART
	NRF_SAADC->ENABLE = 0;	// disable ADC
	NRF_PWM0->ENABLE = 0;	// disable all pwm instance
	NRF_PWM1->ENABLE = 0;
	NRF_PWM2->ENABLE = 0;
	NRF_TWI0->ENABLE = 0;
	NRF_TWIM0->ENABLE = 0; // disable TWI Master
	NRF_TWIS0->ENABLE = 0; // disable TWI Slave
	NRF_TWIM1->ENABLE = 0; // disable TWI Master
	NRF_TWIS1->ENABLE = 0; // disable TWI Slave

	NRF_SPI0->ENABLE = 0; // disable SPI
	NRF_SPI1->ENABLE = 0; // disable SPI
	NRF_SPI2->ENABLE = 0; // disable SPI
	NRF_POWER->TASKS_LOWPWR = 1;
	// NRF_CLOCK->TASKS_LFCLKSTART = 1;

	unsigned int key = irq_lock();
	k_cpu_atomic_idle(key);
	// Enter System ON sleep mode
	__SEV();
	__WFE();
	__WFE();

	return 0;
}
